package core;

import exceptions.IllegalMoveException;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import network.*;
import setup.BoardBuilder;
import setup.ProceduralBoardBuilder;
import tools.Iterators;

/**

 @author reidl
 */
public class GameServer implements Runnable {

    private boolean slowmo = false;
    private LufBoard board = null;
    private boolean isRunning = false;
    private boolean isPlaying = false;
    private BoardBuilder boardBuilder;
    private int timeout = 1000;
    private int knockPort = -1;
    private int numPlayers = 0;
    private int currentPlayer = 0;
    private ArrayList<Player> players = new ArrayList<Player>();
    private ArrayList<Spectator> spectators = new ArrayList<Spectator>();
    private GameServerLog broadcastLog = new GameServerLog();
    private int delay;
    private boolean noKick = false;

    public GameServer() {
    }

    public GameServer setBoardFactory(BoardBuilder boardBuilder) {
        this.boardBuilder = boardBuilder;
        return this;
    }

    public GameServer setPort(int port) {
        this.knockPort = port;
        return this;
    }

    public GameServer setLogfile(String path) {
        broadcastLog.setLogFile(path);
        return this;
    }

    public GameServer setNumPlayers(int numPlayers) {
        // Game starts as soon as this many players have connected
        // 0 denotes an arbitrary numbers (start after timeout?)
        this.numPlayers = numPlayers;
        return this;
    }

    public GameServer setDelay(int delay) {
        if (delay > 0) {
            this.delay = delay;
            this.slowmo = true;
        }
        return this;
    }

    private void log(Object msg) {
        Logger logger = Logger.getLogger("Server");
        logger.log(Level.INFO, "Server: {0}", msg.toString());
    }
    
    private void error(Object msg) {
        Logger logger = Logger.getLogger("Server");
        logger.log(Level.SEVERE, "Server: {0}", msg.toString());
    }    

    public void start() {
        assert (!isRunning);
        isRunning = true;
        players.clear();
        log("Starting game server with config\n"
                + "    numPlayers: " + numPlayers + "\n"
                + "    board builder: " + boardBuilder + "\n"
                + "    timeout: " + timeout + "\n"
                + "    delay: " + delay + "\n"
                + "    no kick: " + noKick + "\n\n");

        KnockServer.create(this, knockPort);
    }

    @Override
    public void run() {
        // Give clients some time
        try {
            synchronized (this) {
                wait(1000);
            }
        } catch (InterruptedException ex) {
            error(ex);
        }

        log("The game is on!");

        // Main game loop
        boolean gameFinished = false;
        while (!gameFinished) {
            if (slowmo) {
                try {
                    synchronized (this) {
                        wait(this.delay);
                    }
                } catch (InterruptedException ex) {
                    error(ex);
                }
            }
            // Debug output
            StringBuilder b = new StringBuilder("Player positions are ");
            for (Player player : players) {
                Position pos = board.getPlayerPosition(player.id);
                b.append(player).append(" ").append(pos).append(", ");
            }
            log(b.toString());

            Player movingPlayer = players.get(currentPlayer);

            MessageYourTurn msgYourTurn = new MessageYourTurn();
            msgYourTurn.timeout = timeout;
            log("Player " + movingPlayer);

            MessageMove msgMove = null;
            LufBoard boardCopy = board.copy();
            try {
                // Send client message
                msgYourTurn.write(movingPlayer.getStream());
                if (noKick) {
                    movingPlayer.setNoTimeout();
                } else {
                    movingPlayer.setTimeout((int) msgYourTurn.timeout);
                }
                msgMove = movingPlayer.readMoveMessage();

                log("Got move message " + msgMove);

                board.insertTile(msgMove.getTileMove());
                board.movePlayer(movingPlayer.id, msgMove.getPlayerMove());
            } catch (IllegalMoveException ex) {
                kickPlayer(movingPlayer, NetworkConstants.REASON_ILLEGAL_MOVE);
                board = boardCopy; // Reset move
                error(ex);
            } catch (MalformedMessageException ex) {
                kickPlayer(movingPlayer, NetworkConstants.REASON_MALFORMED_MESSAGE);
                error(ex);
            } catch (SocketTimeoutException ex) {
                kickPlayer(movingPlayer, NetworkConstants.REASON_MOVE_TIMEOUT);
                error(ex);
            } catch (IOException ex) {
                kickPlayer(movingPlayer, NetworkConstants.REASON_NETWORK_ERROR);
                error(ex);
            }

            ArrayList<Integer> winners = new ArrayList<Integer>();
            if (msgMove != null) {
                log("Move " + msgMove.getTileMove() + ", " + msgMove.getPlayerMove());

                // Augment message with collection events
                msgMove.clearCollectionEvents();
                for (Player player : players) {
                    Set<Integer> collected = board.collectObjectives(player.id);
                    if (!collected.isEmpty()) {
                        log("Player " + player.name + " collected objectives " + collected);
                        msgMove.addCollectionEvents(player.id, collected);
                        if (board.getPlayerObjectives(player.id).isEmpty()) {
                            winners.add(player.id);
                        }
                    }
                }

                broadCast(msgMove);
            }

            if (numPlayers == 0 || !winners.isEmpty()) {
                MessageGameEnd msgGameEnd = new MessageGameEnd();
                msgGameEnd.setWinners(winners);
                broadCast(msgGameEnd);
                gameFinished = true;
            } else {
                currentPlayer = (currentPlayer + 1) % numPlayers;
            }
        }

        log("Game ended.");
        // System.exit(0); // For now.
    }

    public void acceptClient(String name, ServerSocket socket) {
        if (isPlaying) {
            return;
        }

        log("Client " + name + " connected to " + socket.getLocalPort());
        try {
            players.add(new Player(name, players.size(), socket.accept()));
        } catch (IOException ex) {
            error(ex);
        }

        if (players.size() == numPlayers) {
            startGame();
        } else {
            log("Got " + players.size() + " of " + numPlayers + " players");
        }
    }

    public void acceptSpectator(ServerSocket socket) {
        log("Spectator connected to " + socket.getLocalPort());
        try {
            Spectator spectator = new Spectator(socket.accept());
            spectators.add(spectator);
            broadcastLog.replay(spectator);
        } catch (IOException ex) {
            error(ex);
        }
    }

    private void kickPlayer(Player player, byte reason) {
        log("Kicking player " + player);

        // First broadcast so we know even the kicked player
        // gets the message
        MessageKickPlayer msgKickPlayer = new MessageKickPlayer();
        msgKickPlayer.playerId = player.id;
        msgKickPlayer.reason = reason;
        broadCast(msgKickPlayer);

        // Remove player
        int index = players.indexOf(player);
        assert (index != -1);

        players.remove(player);
        numPlayers = players.size();

        if (currentPlayer > index) {
            currentPlayer--;
        }
    }

    private void broadCast(Message message) {
        broadcastLog.logMessage(message);

        HashSet<Spectator> remove = new HashSet<Spectator>();
        for (Spectator spectator : Iterators.join(players, spectators)) {
            try {
                message.write(spectator.getStream());
            } catch (IOException ex) {
                remove.add(spectator);
                error("Failed for " + spectator + "\n " + ex);
            }
        }

        // Spectators are just removed
        spectators.removeAll(remove);

        // TODO: kick players.
    }

    private void startGame() {
        // Mix up player order
        Collections.shuffle(players);

        // Create board 
        board = boardBuilder.setNumPlayers(players.size()).build();

        MessageGameStart msgGameStart = MessageGameStart.fromBoard(board);
        MessagePlayerInfo msgPlayerInfo = new MessagePlayerInfo(numPlayers, board.getPlayerObjectives());
        for (Player player : players) {
            msgPlayerInfo.setPlayerName(player.id, player.name);
        }

        log("Broadcasting game start message");

        for (Spectator spectator : spectators) {
            log("Sending player info to spectator @" + spectator.socket.getPort());
            try {
                // TODO: Write in specs
                msgPlayerInfo.write(spectator.getStream());
                msgGameStart.write(spectator.getStream());
            } catch (IOException ex) {
                error(ex);
            }
        }        

        // Write private part of messages and submit to players
        for (Player player : players) {
            msgGameStart.setPlayerId(player.id);
            msgGameStart.setPlayerObjectives(board.getPlayerObjectives(player.id));
            try {
                log("Sending game start message to " + player.name);
                msgGameStart.write(player.getStream());
            } catch (IOException ex) {
                kickPlayer(player, NetworkConstants.REASON_NETWORK_ERROR);
                error(ex);
            }
        }

        broadcastLog.logMessage(msgPlayerInfo);
        broadcastLog.logMessage(msgGameStart);
        

        // Start main loop
        isPlaying = true;
        (new Thread(this)).start();
    }

    public GameServer setTimeout(int t) {
        this.timeout = t;
        return this;
    }

    public GameServer setNoKick(boolean noKick) {
        this.noKick = noKick;
        return this;
    }

    protected static class Player extends Spectator {

        public final String name;
        private int id;

        public Player(String name, int id, Socket socket) {
            super(socket);
            this.name = name;
            this.id = id;
        }

        public MessageMove readMoveMessage() throws IOException, MalformedMessageException {
            DataInputStream in = new DataInputStream(socket.getInputStream());
            byte header = in.readByte();
            if (header != NetworkConstants.MSG_MOVE) {
                throw new MalformedMessageException();
            }
            return MessageMove.parse(in);
        }

        public void setTimeout(int millis) throws SocketException {
            socket.setSoTimeout(millis);
        }

        public String toString() {
            return name + "(" + id + ")";
        }

        private void setNoTimeout() throws SocketException {
            socket.setSoTimeout(0); // See Javadoc of socket
        }
    }

    protected static class Spectator {

        protected final Socket socket;

        public Spectator(Socket socket) {
            this.socket = socket;
        }

        public DataOutputStream getStream() throws IOException {
            if (socket.isClosed()) {
                System.err.println("Socket was closed for " + toString());
            }
            return new DataOutputStream(socket.getOutputStream());
        }

        public String toString() {
            return "(spectator)";
        }
    }
}
