package core;

import client.LufClient;
import java.io.IOException;
import java.util.logging.FileHandler;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.logging.SimpleFormatter;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;
import org.apache.commons.cli.*;
import setup.BoardBuilder;
import setup.FileBoardBuilder;
import setup.ProceduralBoardBuilder;
import setup.RandomBoardBuilder;

public class Main {

    private static Options options;

    private static void log(Object msg) {
        Logger logger = Logger.getLogger("Main");
        logger.log(Level.INFO, "Main: {0}", msg.toString());
    }

    private static void initializeOptions() {
        // create Options object
        Main.options = new Options();

        // add options
        Main.options.addOption(
                new Option("p", "knockPort", true, "Port for KnockServer to listen on"));
        Main.options.addOption(
                new Option("n", "players", true, "Number of Players to wait for"));
        Main.options.addOption(
                new Option("t", "timeout", true, "Timeout for each turn in msecs (server only)"));
        Main.options.addOption(
                new Option("s", "boardSize", true,
                "Size of the board in format width*height, eg. 10*10 (server only)"));
        Main.options.addOption(
                new Option("m", "map", true, "Load supplied labyrinth map file"));
        Main.options.addOption(
                new Option("d", "delay", true, "Delay between each turn (server only)"));
        Main.options.addOption(
                new Option("k", "noKick", false, "Do not kick players for timeouts (server only)"));

        Main.options.addOption(
                new Option("l", "logFile", true, "File to write logs to (server only)"));
        Main.options.addOption(
                new Option("g", "gameLogFile", true, "File to write game logs to (server only)"));        
        Main.options.addOption(
                new Option("c", "client", false, "Start in client mode, not server"));
        Main.options.addOption(
                new Option("a", "name", true, "Client Player Name (client only)"));
        Main.options.addOption(
                new Option("D", "demo", false, "Start in demo mode, not server"));
    }

    public static void main(String[] args) throws IOException {

        Main.initializeOptions();

        CommandLineParser clp = new GnuParser();

        try {
            // parse the command line arguments
            CommandLine line = clp.parse(Main.options, args);
            
            String logFile = parseStringOption(true, "logFile", line, null);
            if( logFile != null ) {
                FileHandler handler = new FileHandler(logFile);
                handler.setFormatter(new SimpleFormatter());
                
                Logger.getLogger("").addHandler(handler);
            }

            // Branch client <-> server <-> demo
            if (line.hasOption("client")) {
                // Client mode
                log("Starting in Client Mode.");
                // the client requires knockPort and name
                int knockPort = Main.parseIntOption(false, "knockPort", line, -1);
                String name = null;

                name = Main.parseStringOption(false, "name", line, null);

                new LufClient(name, knockPort).start();
            } else if (line.hasOption("demo")) {
                // Demo mode
                int port = 5555;
                int numClients = 4;
                GameServer server = new GameServer();
                ProceduralBoardBuilder builder = new RandomBoardBuilder().setWidth(8).setHeight(8);

                String gameLogFile = Main.parseStringOption(true, "gameLogFile", line, null);
                if (gameLogFile != null) {
                    server.setLogfile(gameLogFile);
                }

                server.setBoardFactory(builder).setPort(port).setNumPlayers(numClients).setTimeout(1000).start();

                new LufClient("Olive", port).start();
                new LufClient("Magenta", port).start();
                new LufClient("Teal", port).start();
                new LufClient("Raspberry", port).start();
            } else {
                // Server mode
                log("Starting in Server Mode.");

                // (" --knockPort=1234 -n=2 --boardSize=25*35 --timeout=5000");
                int knockPort;
                int numPlayers;
                int boardHeight = 10;
                int boardWidth = 10;
                int timeout = 1000;
                int delay = -1;
                boolean noKick = false;

                String gameLogFile = null;

                knockPort = Main.parseIntOption(false, "knockPort", line, -1);
                numPlayers = Main.parseIntOption(false, "players", line, -1);

                timeout = Main.parseIntOption(true, "timeout", line, 1000);
                delay = Main.parseIntOption(true, "delay", line, -1);
                noKick = line.hasOption("noKick");

                gameLogFile = Main.parseStringOption(true, "gameLogFile", line, null);

                BoardBuilder builder = null;

                if (line.hasOption("map")) {
                    String file = Main.parseStringOption(false, "map", line, null);
                    builder = new FileBoardBuilder(file);
                } else {
                    if (line.hasOption("boardSize")) {
                        String boardSize = line.getOptionValue("boardSize");

                        int p = boardSize.indexOf("*");
                        if (p != -1) {
                            String swidth = boardSize.substring(0, p);
                            String sheight = boardSize.substring(p + 1);
                            try {
                                boardWidth = Integer.parseInt(swidth);
                                boardHeight = Integer.parseInt(sheight);
                            } catch (NumberFormatException exp) {
                                System.out
                                        .println("Incorrect parameter string for boardSize: Must be in format width*height");
                                Main.printHelpAndQuit();
                            }
                        } else {
                            log("Incorrect parameter string for boardSize: Must be in format width*height");
                            Main.printHelpAndQuit();
                        }
                    }

                    builder = new RandomBoardBuilder().setWidth(boardWidth).setHeight(boardHeight);
                }

                GameServer server = new GameServer();

                if (gameLogFile != null) {
                    server.setLogfile(gameLogFile);
                }

                server.setBoardFactory(builder).setPort(knockPort).setNumPlayers(numPlayers).setTimeout(timeout)
                        .setDelay(delay).setNoKick(noKick).start();
            }
        } catch (ParseException exp) {
            log("Parsing arguments failed: " + exp.getMessage());
            Main.printHelpAndQuit();
        } catch (NumberFormatException ex) {
            log("Parsing arguments failed (not a number): " + ex.getMessage());
            Main.printHelpAndQuit();
        }
    }

    private static int parseIntOption(boolean isOptional, String name, CommandLine line, int defaultValue) {
        int result = defaultValue;
        if (line.hasOption(name)) {
            try {
                result = Integer.parseInt(line.getOptionValue(name));
            } catch (NumberFormatException exp) {
                log("Parameter " + name + " was not a number!");
                Main.printHelpAndQuit();
            }
        } else if (!isOptional) {
            log("Required parameter " + name + " was missing.");
            Main.printHelpAndQuit();
        }

        return result;
    }

    private static String parseStringOption(boolean isOptional, String name, CommandLine line, String defaultValue) {
        String result = defaultValue;
        if (line.hasOption(name)) {
            result = line.getOptionValue(name);
        } else if (!isOptional) {
            log("Required parameter " + name + " was missing.");
            Main.printHelpAndQuit();
        }

        return result;
    }

    private static void printHelpAndQuit() {
        System.out.println("Game server & test client v1.5. Usage: ");

        HelpFormatter helpFormatter = new HelpFormatter();
        helpFormatter.printHelp("java -jar labyrinth.jar", Main.options);

        System.out.println("Server example cmd:");
        System.out.println(" --knockPort=1234 -n=2 --boardSize=25*35 --timeout=5000");
        System.out.println("Client example cmd:");
        System.out.println(" --client --knockPort=1234 --name=blub");
        System.out.println("Demo example cmd:");
        System.out.println(" --gameLogFile=file.ext");

        System.out.println("Terminating. Goodbye...");
        System.exit(-1);
    }
}
