package setup;

import core.LufBoard;
import core.Position;
import core.TileType;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.management.RuntimeErrorException;
import tools.TileTools;
import tools.TileTools.ExtendedTileType;

/**
 Important: (-1,-1) denotes free tile position
 Board format (empty lines and lines beginning with # are ignored)
 width w
 height h
 num players n
 player positions x1,y1,x2,y2,...,xn,yn
 objectives player 1 x1,y1,...
 objectives player 2 x1,y1,...
 ...
 objectives player n x1,y1,...
 free tile (see below for symbols)
 board matrix: h lines of length w using symbols
 ╳ ═ ║ ╝ ╚ ╔ ╗ ╩ ╠ ╦ ╣ ╬  moveable
 ✖ ━ ┃ ┛ ┗ ┏ ┓ ┻ ┣ ┳ ┫ ╋  unmoveable

 @author reidl
 */
public class FileBoardBuilder extends BoardBuilder {

    private final String file;

    public FileBoardBuilder(String file) {
        this.file = file;
    }

    @Override
    public LufBoard build() {
        File f = new File(file, "UTF-8");
        LufBoard res = null;
        try {
            BufferedReader reader = new BufferedReader(new FileReader(file));

            // Read basic data
            int width = readInt(reader);
            int height = readInt(reader);
            res = new LufBoard(width, height);
            super.addVisitorsToBoard(res);
            
            ExtendedTileType[][] transTiles = new ExtendedTileType[height][width];

            if (width <= 0 || height <= 0) {
                abort("Illegal board dimension: " + width + " x " + height);
            }

            int numPlayers = readInt(reader);
            
            if( this.numPlayers != -1 && this.numPlayers != numPlayers) {
                abort("Server expected differnet number of players then specified in the map file.");
            }
            
            Position[] playerPos = readPosArray(reader);

            if (playerPos.length != numPlayers) {
                abort("Number of player positios does not match number of players.");
            }

            // Read objectives
            ArrayList<Position> objectivePositions = new ArrayList<Position>();
            ArrayList<ArrayList<Integer>> playerObjectives = new ArrayList<ArrayList<Integer>>();
            for (int i = 0; i < numPlayers; i++) {
                Position[] objectives = readPosArray(reader);

                int offset = objectivePositions.size();
                ArrayList<Integer> objectiveIndices = new ArrayList<Integer>();
                for (int j = 0; j < objectives.length; j++) {
                    objectiveIndices.add(j + offset);
                }
                playerObjectives.add(objectiveIndices);

                objectivePositions.addAll(Arrays.asList(objectives));
            }

            // Read free tile
            TileType freeTile = readTiles(reader)[0].type;

            // Read board matrix
            for (int y = 0; y < height; y++) {
                ExtendedTileType[] row = readTiles(reader);
                if(row.length != width) {
                    abort("Row " + y + " has wrong width.");
                }
                
                transTiles[y] = row;
            }
            
            ExtendedTileType[][] tiles = TileTools.transpose(transTiles);
                        
            assert(tiles.length == width && tiles[0].length == height);
            
            res.setBoard(TileTools.downcast(tiles), 
                         freeTile, objectivePositions.toArray(new Position[0]),
                         playerPos, 
                         TileTools.extractUnmoveable(tiles));
            
            for (int i = 0; i < numPlayers; i++) {
                res.setPlayerObjectives(i, playerObjectives.get(i));
            }
        } catch (FileNotFoundException ex) {
            throw new RuntimeException("Board file does not exist or is not readable: " + file);
        } catch (IOException ex) {
            throw new RuntimeException("Board file not readable or malformed: " + file);
        } catch (NumberFormatException ex) {
            throw new RuntimeException("Board file malformed: " + file);
        }

        return res;
    }

    private String nextLine(BufferedReader reader) throws IOException {
        String line = "";
        while (line.length() == 0) {
            line = reader.readLine().trim();
            int pos = line.indexOf("#");
            if( pos != -1 ) {
                line = line.substring(0,pos);
            }
        }
        return line;
    }

    private int readInt(BufferedReader reader) throws IOException {
        String line = nextLine(reader);
        return Integer.parseInt(line);
    }

    private int[] readIntArray(BufferedReader reader) throws IOException {
        String line = nextLine(reader);

        String[] tokens = line.split(",");
        int[] res = new int[tokens.length];
        for (int i = 0; i < res.length; i++) {
            res[i] = Integer.parseInt(tokens[i].trim());
        }


        return res;
    }

    private Position[] readPosArray(BufferedReader reader) throws IOException {
        int[] raw = readIntArray(reader);
        if ((raw.length % 2) != 0) {
            abort("Tried to read position array, but length is odd.");
        }

        Position[] res = new Position[raw.length / 2];
        for (int i = 0; i < res.length; i++) {
            res[i] = new Position(raw[2 * i], raw[2 * i + 1]);
        }

        return res;
    }

    private void abort(String message) throws RuntimeException {
        throw new RuntimeException(message);
    }

    private ExtendedTileType[] readTiles(BufferedReader reader) throws IOException {
        String line = nextLine(reader);
        int len = line.length();
        ArrayList<ExtendedTileType> row = new ArrayList<ExtendedTileType>();
        for (int i = 0; i < len; i++) {
            row.add(TileTools.fromChar(line.charAt(i)));
        }
        return row.toArray(new ExtendedTileType[0]);
    }
}
