package tools;

import core.Position;
import core.TileType;
import static core.TileType.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;

/**

 @author ubik
 */
public class TileTools {

    private static HashSet<TileType> withExitRight;
    private static HashSet<TileType> withExitLeft;
    private static HashSet<TileType> withExitTop;
    private static HashSet<TileType> withExitBottom;
    private static HashMap<Character, TileType> charToType;
    private static HashSet<Character> charUnmoveable;
    private static TileType[][] traces;

    static {
        withExitRight = new HashSet<TileType>(Arrays.asList(
                new TileType[]{
                    LR, LRB, LTR, LTRB, RB, TR, TRB
                }));

        withExitLeft = new HashSet<TileType>(Arrays.asList(
                new TileType[]{
                    LB, LR, LRB, LT, LTB, LTR, LTRB
                }));

        withExitTop = new HashSet<TileType>(Arrays.asList(
                new TileType[]{
                    LT, LTB, LTR, LTRB, TB, TR, TRB
                }));

        withExitBottom = new HashSet<TileType>(Arrays.asList(
                new TileType[]{
                    LB, LRB, LTB, LTRB, RB, TB, TRB
                }));

        traces = new TileType[TileType.values().length][];
        traces[BLANK.ordinal()] = new TileType[]{BLANK};
        traces[LTRB.ordinal()] = new TileType[]{LTRB};
        traces[TB.ordinal()] = new TileType[]{TB, LR};
        traces[LR.ordinal()] = new TileType[]{LR, TB};

        traces[LB.ordinal()] = new TileType[]{LB, LT, TR, RB};
        traces[LT.ordinal()] = new TileType[]{LT, TR, RB, LB};
        traces[TR.ordinal()] = new TileType[]{TR, RB, LB, LT};
        traces[RB.ordinal()] = new TileType[]{RB, LB, LT, TR};

        traces[LTR.ordinal()] = new TileType[]{LTR, TRB, LRB, LTB};
        traces[TRB.ordinal()] = new TileType[]{TRB, LRB, LTB, LTR};
        traces[LRB.ordinal()] = new TileType[]{LRB, LTB, LTR, TRB};
        traces[LTB.ordinal()] = new TileType[]{LTB, LTR, TRB, LRB};

        TileType[] values = TileType.values();
        charToType = new HashMap<Character, TileType>();
        for (int i = 0; i < values.length; i++) {
            TileType type = values[i];
            charToType.put(toCharacter(type, true), type);
            charToType.put(toCharacter(type, false), type);
            assert (traces[type.ordinal()][0] == type);
        }

        charUnmoveable = new HashSet<Character>();
        charUnmoveable.addAll(Arrays.asList(new Character[]{'✖', '━', '┃', '┛', '┗', '┏', '┓', '┻', '┣', '┳', '┫', '╋'}));
    }

    public static boolean hasExitRight(TileType type) {
        return withExitRight.contains(type);
    }

    public static boolean hasExitLeft(TileType type) {
        return withExitLeft.contains(type);
    }

    public static boolean hasExitTop(TileType type) {
        return withExitTop.contains(type);
    }

    public static boolean hasExitBottom(TileType type) {
        return withExitBottom.contains(type);
    }

    public static Character toCharacter(TileType type, boolean moveable) {
        switch (type) {
            case BLANK:
                return moveable ? '╳' : '✖';
            case LR:
                return moveable ? '═' : '━';
            case TB:
                return moveable ? '║' : '┃';
            case LT:
                return moveable ? '╝' : '┛';
            case TR:
                return moveable ? '╚' : '┗';
            case RB:
                return moveable ? '╔' : '┏';
            case LB:
                return moveable ? '╗' : '┓';
            case LTR:
                return moveable ? '╩' : '┻';
            case TRB:
                return moveable ? '╠' : '┣';
            case LRB:
                return moveable ? '╦' : '┳';
            case LTB:
                return moveable ? '╣' : '┫';
            case LTRB:
                return moveable ? '╬' : '╋';
            default:
                throw new AssertionError(type.name());
        }
    }

    public static ExtendedTileType fromChar(char c) {
        return new ExtendedTileType(charToType.get(c), charUnmoveable.contains(c));
    }

    public static TileType[] getTrace(TileType freeTile) {
        return traces[freeTile.ordinal()];
    }

    public static TileType[][] transpose(TileType[][] matrix) {
        TileType[][] res = new TileType[matrix[0].length][matrix.length];

        for (int x = 0; x < matrix.length; x++) {
            for (int y = 0; y < matrix[0].length; y++) {
                res[y][x] = matrix[x][y];
            }
        }

        return res;
    }

    public static ExtendedTileType[][] transpose(ExtendedTileType[][] matrix) {
        ExtendedTileType[][] res = new ExtendedTileType[matrix[0].length][matrix.length];

        for (int x = 0; x < matrix.length; x++) {
            for (int y = 0; y < matrix[0].length; y++) {
                res[y][x] = matrix[x][y];
            }
        }

        return res;
    }

    public static TileType[][] downcast(ExtendedTileType[][] matrix) {
        int width = matrix.length;
        int height = matrix[0].length;
        TileType[][] res = new TileType[width][height];

        for (int x = 0; x < width; x++) {
            for (int y = 0; y < height; y++) {
                res[x][y] = matrix[x][y].type;
            }
        }

        return res;
    }

    public static Position[] extractUnmoveable(ExtendedTileType[][] matrix) {
        int width = matrix.length;
        int height = matrix[0].length;
        ArrayList<Position> res = new ArrayList<Position>();

        for (int x = 0; x < width; x++) {
            for (int y = 0; y < height; y++) {
                if (matrix[x][y].unmovable) {
                    res.add(Position.create(x, y));
                }
            }
        }

        return res.toArray(new Position[0]);
    }

    /**
     Returns the number of clockwise 90 degree-rotations needed to
     transform the tile type 'source' into the type 'target'

     @param source
     @param target
     @return
     */
    public static int calculateRotation(TileType source, TileType target) {
        TileType[] trace = getTrace(source);
        for (int i = 0; i < trace.length; i++) {
            if (trace[i] == target) {
                return i;
            }
        }
        throw new RuntimeException("Tile types not of the same rotational group" + source + ", " + target);
    }

    /**
     * Rotates the tile once by 90 degree, clockwise.
     * @param type
     * @return 
     */
    public static TileType rotate(TileType type) {
        TileType[] trace = getTrace(type);
        return trace[1%trace.length];
	}

    public static class ExtendedTileType {

        public final TileType type;
        public final boolean unmovable;

        public ExtendedTileType(TileType type, boolean unmovable) {
            this.type = type;
            this.unmovable = unmovable;
        }
    }
}
